@@ -10,10 +10,10 @@ from cftool.types import tensor_dict_type
 from .constants import LV1_ALPHA_KEY
 from .constants import LV1_RAW_ALPHA_KEY
 from ...bases import CascadeMixin
-from ....protocol import TrainerState
+from ....schema import TrainerState
 from ....constants import INPUT_KEY
 from ....constants import PREDICTIONS_KEY
-from ...protocols.cv import ImageTranslatorMixin
+from ...schemas.cv import ImageTranslatorMixin
 from ....misc.toolkit import interpolate
 from ....misc.internal_.register import register_module
 
